#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from migen import *

from litespi.core.mmap import LiteSPIMMAP
from litespi.common import *
from litespi.opcodes import SpiNorFlashOpCodes as Codes
from litespi.spi_nor_flash_module import SpiNorFlashModule
from litespi.ids import SpiNorFlashManufacturerIDs


class TestSPIMMAP(unittest.TestCase):
    class DummyChip(SpiNorFlashModule):

        manufacturer_id = SpiNorFlashManufacturerIDs.NONJEDEC
        device_id = 0x0204
        name = "dummychip1"

        total_size  =    2097152   # bytes
        page_size   =        256   # bytes
        total_pages =       8192

        supported_opcodes = [
            Codes.READ_1_1_1,
            Codes.PP_1_1_1
        ]
        dummy_bits = 8

    def test_spi_mmap_core_syntax(self):
        spi_mmap = LiteSPIMMAP(flash=self.DummyChip(Codes.READ_1_1_1, []))

    def test_spi_mmap_read_test(self):
        opcode = Codes.READ_1_1_1
        dut = LiteSPIMMAP(flash=self.DummyChip(opcode, []))

        def wb_gen(dut, addr, data):
            dut.data_ok = 0

            yield dut.bus.adr.eq(addr)
            yield dut.bus.we.eq(0)
            yield dut.bus.cyc.eq(1)
            yield dut.bus.stb.eq(1)

            while (yield dut.bus.ack) == 0:
                yield
            print((yield dut.bus.dat_r))
            if (yield dut.bus.dat_r) == data:
                dut.data_ok = 1

        def phy_gen(dut, addr, data):
            dut.addr_ok = 0
            dut.opcode_ok = 0
            dut.cmd_ok = 0
            dut.no_dummy = 0
            yield dut.sink.valid.eq(0)
            yield dut.source.ready.eq(1)

            while (yield dut.source.valid) == 0:
                yield


            # READ CMD
            if (yield dut.source.data) == opcode.code: # cmd ok
                dut.opcode_ok = 1

            yield
            yield dut.sink.valid.eq(1)
            while (yield dut.source.valid) == 0:
                yield
            yield dut.sink.valid.eq(0)

            # READ ADDR
            if (yield dut.source.data) == (addr<<2): # address cmd
                dut.addr_ok = 1

            yield
            yield dut.sink.valid.eq(1)
            while (yield dut.source.valid) == 0:
                yield
            yield dut.sink.valid.eq(0)

            # NO DUMMY, mask should be 0 cause we attept to read data
            if (yield dut.source.mask) == 0:
                dut.no_dummy = 1

            # SEND DATA
            yield dut.source.ready.eq(0)
            yield
            yield dut.sink.data.eq(data)
            yield dut.sink.valid.eq(1)

            while (yield dut.sink.ready) == 0:
                yield

            yield
            yield dut.sink.valid.eq(0)
            yield
        addr = 0xcafe
        data = 0xdeadbeef

        run_simulation(dut, [wb_gen(dut, addr, data), phy_gen(dut, addr, data)])
        self.assertEqual(dut.data_ok, 1)
        self.assertEqual(dut.addr_ok, 1)
        self.assertEqual(dut.opcode_ok, 1)
        self.assertEqual(dut.no_dummy, 1)



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.integration.doc import AutoDoc
from litex.soc.interconnect import wishbone, stream
from litex.soc.interconnect.csr import *

from litespi.common import *
from litespi.crossbar import LiteSPICrossbar
from litespi.core.master import LiteSPIMaster
from litespi.core.mmap import LiteSPIMMAP


class LiteSPICore(Module):
    def __init__(self):
        self.source = stream.Endpoint(spi_core2phy_layout)
        self.sink   = stream.Endpoint(spi_phy2core_layout)
        self.cs     = Signal()


class LiteSPI(Module, AutoCSR, AutoDoc):
    """SPI Controller wrapper.

    The ``LiteSPI`` class provides a wrapper that can instantiate both ``LiteSPIMMAP`` and ``LiteSPIMaster`` and connect them to the PHY.

    Both options can be used at the same time with help of ``mux_sel`` register which allows to share access to PHY via crossbar.

    Parameters
    ----------
    phy : Module
        Module or object that contains PHY stream interfaces and a cs signal to connect the ``LiteSPICore`` to.

    clk_freq : int
        Frequency of a clock connected to LiteSPI.

    clock_domain : str
        Name of LiteSPI clock domain.

    with_mmap : bool
        Enables memory-mapped SPI flash controller.

    with_master : bool
        Enables register-operated SPI master controller.

    mmap_endianness : string
        If endianness is set to ``small`` then byte order of each 32-bit word comming MMAP core will be reversed.

    Attributes
    ----------
    bus : Interface(), out
        Wishbone interface for memory-mapped flash access.
    """

    def __init__(self, phy, clock_domain="sys",
        with_mmap=True, mmap_endianness="big",
        with_master=True, master_tx_fifo_depth=1, master_rx_fifo_depth=1,
        with_csr=True):

        self.submodules.crossbar = crossbar = LiteSPICrossbar(clock_domain)
        self.comb += phy.cs.eq(crossbar.cs)

        if with_mmap:
            self.submodules.mmap = mmap = LiteSPIMMAP(flash=phy.flash,
                                                      endianness=mmap_endianness,
                                                      with_csr=with_csr)
            port_mmap = crossbar.get_port(mmap.cs)
            self.bus = mmap.bus
            self.comb += [
                port_mmap.source.connect(mmap.sink),
                mmap.source.connect(port_mmap.sink),
            ]
            if hasattr(phy, "dummy_bits"):
                self.comb += phy.dummy_bits.eq(mmap._spi_dummy_bits)
        if with_master:
            self.submodules.master = master = LiteSPIMaster(
                tx_fifo_depth = master_tx_fifo_depth,
                rx_fifo_depth = master_rx_fifo_depth)
            port_master = crossbar.get_port(master.cs)
            self.comb += [
                port_master.source.connect(master.sink),
                master.source.connect(port_master.sink),
            ]

        if clock_domain != "sys":
            self.comb += [
                crossbar.tx_cdc.source.connect(phy.sink),
                phy.source.connect(crossbar.rx_cdc.sink),
            ]
        else:
            self.comb += [
                crossbar.master.source.connect(phy.sink),
                phy.source.connect(crossbar.master.sink),
            ]



#
# This file is part of LiteSPI
#
# Copyright (c) 2015 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from collections import OrderedDict

from migen import *
from migen.genlib.roundrobin import RoundRobin
from litespi.common import *

from litex.soc.interconnect import stream


class LiteSPIMasterPort:
    def __init__(self):
        self.source = stream.Endpoint(spi_core2phy_layout)
        self.sink   = stream.Endpoint(spi_phy2core_layout)


class LiteSPISlavePort:
    def __init__(self):
        self.source = stream.Endpoint(spi_phy2core_layout)
        self.sink   = stream.Endpoint(spi_core2phy_layout)


class LiteSPICrossbar(Module):
    def __init__(self, cd):
        self.cd     = cd
        self.users  = []
        self.master = LiteSPIMasterPort()
        if cd != "sys":
            rx_cdc = stream.AsyncFIFO(spi_phy2core_layout, 32, buffered=True)
            tx_cdc = stream.AsyncFIFO(spi_core2phy_layout, 32, buffered=True)
            self.submodules.rx_cdc = ClockDomainsRenamer({"write": cd, "read": "sys"})(rx_cdc)
            self.submodules.tx_cdc = ClockDomainsRenamer({"write": "sys", "read": cd})(tx_cdc)
            self.comb += [
                self.rx_cdc.source.connect(self.master.sink),
                self.master.source.connect(self.tx_cdc.sink),
            ]

        self.cs           = Signal()
        self.user_cs      = []
        self.user_request = []

    def get_port(self, cs, request = None):
        user_port     = LiteSPISlavePort()
        internal_port = LiteSPISlavePort()

        tx_stream = user_port.sink

        self.comb += tx_stream.connect(internal_port.sink)

        rx_stream = internal_port.source

        self.comb += rx_stream.connect(user_port.source)

        if request is None:
            request = Signal()
            self.comb += request.eq(cs)

        self.users.append(internal_port)
        self.user_cs.append(self.cs.eq(cs))
        self.user_request.append(request)

        return user_port

    def do_finalize(self):
        self.submodules.rr = RoundRobin(len(self.users))

        # TX
        self.submodules.tx_mux = tx_mux = stream.Multiplexer(spi_core2phy_layout, len(self.users))

        # RX
        self.submodules.rx_demux = rx_demux = stream.Demultiplexer(spi_phy2core_layout, len(self.users))

        for i, user in enumerate(self.users):
            self.comb += [
                user.sink.connect(getattr(tx_mux, f"sink{i}")),
                getattr(rx_demux, f"source{i}").connect(user.source),
            ]

        self.comb += [
            self.rr.request.eq(Cat(self.user_request)),

            self.tx_mux.source.connect(self.master.source),
            self.tx_mux.sel.eq(self.rr.grant),

            self.master.sink.connect(self.rx_demux.sink),
            self.rx_demux.sel.eq(self.rr.grant),

            Case(self.rr.grant, dict(enumerate(self.user_cs))),
        ]



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.fsm import FSM, NextState

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import *

from litespi.common import *


class LiteSPIMaster(Module, AutoCSR):
    """Generic LiteSPI Master

    The ``LiteSPIMaster`` class provides a generic SPI master that can be controlled using CSRs.

    It supports multiple access modes with help of ``width`` and ``mask`` registers which can be used to configure the PHY into any supported SDR mode (single/dual/quad/octal).

    Parameters
    ----------
    fifo_depth : int
        Depth of the internal TX/RX FIFO.

    cs_width : int
        Number of CS lines to support.

    Attributes
    ----------
    source : Endpoint(spi_phy2core_layout), out
        Data stream.

    sink : Endpoint(spi_core2phy_layout), in
        Control stream.

    cs : Signal(), out
        Slave CS signal.

    """
    def __init__(self, cs_width=1, tx_fifo_depth=1, rx_fifo_depth=1):
        self.sink   = stream.Endpoint(spi_phy2core_layout)
        self.source = stream.Endpoint(spi_core2phy_layout)
        self.cs     = Signal(cs_width)
        assert self.sink.data.nbits == self.source.data.nbits

        self._cs = CSRStorage(cs_width)
        self._phyconfig = CSRStorage(fields=[
            CSRField("len",   size=8, offset=0,  description="SPI Xfer length (in bits)."),
            CSRField("width", size=4, offset=8,  description="SPI Xfer width (1/2/4/8)."),
            CSRField("mask",  size=8, offset=16, description="SPI DQ output enable mask (set bits to ``1`` to enable output drivers on DQ lines)."),
        ], description="SPI PHY settings.")
        self._rxtx = CSR(self.source.data.nbits)
        self._status = CSRStatus(fields=[
            CSRField("tx_ready", size=1, offset=0, description="TX FIFO is not full."),
            CSRField("rx_ready", size=1, offset=1, description="RX FIFO is not empty."),
        ])

        # # #

        # FIFOs.
        tx_fifo = stream.SyncFIFO(spi_core2phy_layout, depth=tx_fifo_depth)
        rx_fifo = stream.SyncFIFO(spi_phy2core_layout, depth=rx_fifo_depth)
        self.submodules += tx_fifo, rx_fifo
        self.comb += self.sink.connect(rx_fifo.sink)
        self.comb += tx_fifo.source.connect(self.source)

        # SPI CS.
        self.comb += self.cs.eq(self._cs.storage)

        # SPI TX (MOSI).
        self.comb += [
            tx_fifo.sink.valid.eq(self._rxtx.re),
            self._status.fields.tx_ready.eq(tx_fifo.sink.ready),
            tx_fifo.sink.data.eq(self._rxtx.r),
            tx_fifo.sink.len.eq(self._phyconfig.fields.len),
            tx_fifo.sink.width.eq(self._phyconfig.fields.width),
            tx_fifo.sink.mask.eq(self._phyconfig.fields.mask),
            tx_fifo.sink.last.eq(1),
        ]

        # SPI RX (MISO).
        self.comb += [
            rx_fifo.source.ready.eq(self._rxtx.we),
            self._status.fields.rx_ready.eq(rx_fifo.source.valid),
            self._rxtx.w.eq(rx_fifo.source.data),
        ]



from litespi.modules.generated_modules import *
from litespi.modules.modules import *
from litespi.spi_nor_flash_module import MetaSizes

def print_modules():
    for name, obj in globals().items():
        if not issubclass(type(obj), MetaSizes):
            continue
        if name == 'SpiNorFlashModule':
            continue
        if name == 'MetaSizes':
            continue
        print(obj.table("page"))



#!/usr/bin/env python3

#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause


from collections import namedtuple


SpiNorFlashOpCode = namedtuple("SpiNorFlashOpCode", "code desc")
_Op = SpiNorFlashOpCode


class SpiNorFlashOpCodes:
    """SPI NOR Flash command opcodes."""

    UNKNOWN            = _Op(0x00, "Unknown")

    # Release Power-Down / Device ID
    WREN               = _Op(0x06, "Write enable")
    RDSR               = _Op(0x05, "Read status register")
    WRSR               = _Op(0x01, "Write status register 1 byte")
    RDSR2              = _Op(0x3f, "Read status register 2")
    WRSR2              = _Op(0x3e, "Write status register 2")

    # Reading op codes
    READ_1_1_1         = _Op(0x03, "Read data bytes (low frequency)")
    READ_1_1_1_FAST    = _Op(0x0b, "Read data bytes (high frequency)")
    READ_1_1_2         = _Op(0x3b, "Read data bytes (Dual Output SPI)")
    READ_1_2_2         = _Op(0xbb, "Read data bytes (Dual I/O SPI)")
    READ_1_1_4         = _Op(0x6b, "Read data bytes (Quad Output SPI)")
    READ_1_4_4         = _Op(0xeb, "Read data bytes (Quad I/O SPI)") # Fast Read Quad
    READ_4_4_4         = _Op(0xeb, "Read data bytes (Quad I/O QPI)")
    READ_1_1_8         = _Op(0x8b, "Read data bytes (Octal Output SPI)")
    READ_1_8_8         = _Op(0xcb, "Read data bytes (Octal I/O SPI)")

    # FIXME: What is this?
    SET_BURST_WRAP     = _Op(0x77, "Set burst with wrap")

    # Programming op codes
    PP_1_1_1           = _Op(0x02, "Page program (up to 256 bytes)")
    PP_1_1_2           = _Op(0xa2, "Dual page program")
    PP_1_1_4           = _Op(0x32, "Quad page program")
    PP_1_4_4           = _Op(0x38, "Quad page program")
    PP_1_1_8           = _Op(0x82, "Octal page program")
    PP_1_8_8           = _Op(0xc2, "Octal page program")

    # Erasing op codes
    SE                 = _Op(0xd8, "Sector erase (usually 64KiB)")

    BE_256             = _Op(0xdb, "Erase 256B block")
    BE_4K              = _Op(0x20, "Erase 4KiB block")
    BE_4K_PMC          = _Op(0xd7, "Erase 4KiB block on PMC chips")
    BE_32K             = _Op(0x52, "Erase 32KiB block")
    BE_ALT1            = _Op(0x50, "Erase ??? block")
    BE_ALT2            = _Op(0x81, "Erase ??? block")

    CHIP_ERASE         = _Op(0xc7, "Erase whole flash chip")
    CHIP_ERASE_ALT     = _Op(0x60, "Erase whole flash chip on some chips")
    CHIP_ERASE_ATMEL   = _Op(0x62, "Erase whole flash chip on Atmel chips")


    # Register opcodes
    RDID               = _Op(0x9f, "Read JEDEC ID")
    RDID_ATMEL         = _Op(0x15, "Read JEDEC ID on Atmel chips")
    RDSFDP             = _Op(0x5a, "Read SFDP")
    RDCR               = _Op(0x35, "Read configuration register")
    RDFSR              = _Op(0x70, "Read flag status register")
    CLFSR              = _Op(0x50, "Clear flag status register")
    RDEAR              = _Op(0xc8, "Read Extended Address Register")
    WREAR              = _Op(0xc5, "Write Extended Address Register")

    # - 10.2.29 Release Power-down / Device ID (ABh)
    # The Release from Power-down / Device ID instruction is a multi-purpose
    # instruction. It can be used to release the device from the power-down
    # state, or obtain the devices electronic identification (ID) number.
    PWRUP_RDID         = _Op(0xab, "Power up flash and read device ID")

    # - 10.2.32 Read Manufacturer / Device ID Quad I/O (94h)
    RDID_1_4_4         = _Op(0x94, "Read manufacturer and device ID (Quad I/O SPI)")
    # - 10.2.33 Read Unique ID Number (4Bh)
    RDUID              = _Op(0x4b, "Read unique ID number")

    # - 10.2.37 Read Security Registers (48h)
    RDSFR              = _Op(0x48, "Read Security Registers")
    # - Set Read Parameters (C0h), modifies behaviour of READ_FAST/READ_1_4_4/READ_FAST_4B commands.
    SET_READ_PARAMS    = _Op(0xc0, "Set read parameters")

    # 4-byte address opcodes - used on Spansion and some Macronix flashes.
    READ_1_1_1_4B      = _Op(0x13, "Read data bytes (low frequency)")
    READ_1_1_1_FAST_4B = _Op(0x0c, "Read data bytes (high frequency)")
    READ_1_1_2_4B      = _Op(0x3c, "Read data bytes (Dual Output SPI)")
    READ_1_2_2_4B      = _Op(0xbc, "Read data bytes (Dual I/O SPI)")
    READ_1_1_4_4B      = _Op(0x6c, "Read data bytes (Quad Output SPI)")
    READ_1_4_4_4B      = _Op(0xec, "Read data bytes (Quad I/O SPI)")
    READ_1_1_8_4B      = _Op(0x7c, "Read data bytes (Octal Output SPI)")
    READ_1_8_8_4B      = _Op(0xcc, "Read data bytes (Octal I/O SPI)")

    PP_1_1_1_4B        = _Op(0x12, "Page program (up to 256 bytes)")
    PP_1_1_4_4B        = _Op(0x34, "Quad page program")
    PP_1_4_4_4B        = _Op(0x3e, "Quad page program")
    PP_1_1_8_4B        = _Op(0x84, "Octal page program")
    PP_1_8_8_4B        = _Op(0x8e, "Octal page program")

    BE_4K_4B           = _Op(0x21, "Erase 4KiB block")
    BE_32K_4B          = _Op(0x5c, "Erase 32KiB block")
    SE_4B              = _Op(0xdc, "Sector erase (usually 64KiB)")

    # Word Read Quad I/O - E7h
    # Address bits (A0) must equal 0
    READ_2ALIGN_1_4_4  = _Op(0xe7, "Aligned 16bit (word) read data bytes (Quad I/O SPI)")
    # Address bits (A3, A2, A1, A0) must equal 0
    READ_8ALIGN_1_4_4  = _Op(0xe3, "Aligned 128bit read data bytes (Quad I/O SPI)")

    # Double Transfer Rate opcodes - defined in JEDEC JESD216B.
    READ_1_1_1_DTR     = _Op(0x0d, "")
    READ_1_2_2_DTR     = _Op(0xbd, "")
    READ_1_4_4_DTR     = _Op(0xed, "")

    READ_1_1_1_DTR_4B  = _Op(0x0e, "")
    READ_1_2_2_DTR_4B  = _Op(0xbe, "")
    READ_1_4_4_DTR_4B  = _Op(0xee, "")

    # Used for SST flashes only.
    BP                 = _Op(0x02, "Byte program")
    WRDI               = _Op(0x04, "Write disable")

    # Used for S3AN flashes only
    XSE                = _Op(0x50, "Sector erase")
    XPP                = _Op(0x82, "Page program")
    XRDSR              = _Op(0xd7, "Read status register")

    # Used for Macronix and Winbond flashes.
    EN4B               = _Op(0xb7, "Enter 4-byte mode")
    EX4B               = _Op(0xe9, "Exit 4-byte mode")

    # Used for Spansion flashes only.
    BRWR               = _Op(0x17, "Bank register write")
    CLSR               = _Op(0x30, "Clear status register 1")

    # Used for Micron flashes only.
    RD_EVCR            = _Op(0x65, "Read EVCR register")
    WD_EVCR            = _Op(0x61, "Write EVCR register")

    # JEDEC "Auto address increment" mode.
    # - AAI supported, but opcode is 0xAF
    AAI_WP             = _Op(0xad, "Auto address increment word program")

    # Opcode aliases
    READ         = READ_1_1_1
    READ_FAST    = READ_1_1_1_FAST
    FAST_READ    = READ_FAST
    READ_4B      = READ_1_1_1_4B
    READ_FAST_4B = READ_1_1_1_FAST_4B
    DQFR         = READ_1_1_2
    DIOFR        = READ_1_2_2
    QOFR         = READ_1_1_4
    QUIFR        = READ_1_4_4
    PP           = PP_1_1_1
    PP_4B        = PP_1_1_1_4B

    # JEDEC Aliases for various op codes
    JEDEC_READ              = READ          # 0x03
    JEDEC_READ_FAST         = READ_FAST     # 0x0b
    JEDEC_READ_4BA          = READ_4B       # 0x13
    JEDEC_READ_4BA_FAST     = READ_FAST_4B  # 0x0c
    JEDEC_READ_EXT_ADDR_REG = RDEAR         # 0xc8
    JEDEC_AAI_WORD_PROGRAM  = AAI_WP        # 0xad

    @property
    def _get_configs(self):
        # Pattern: x_x_x_<FAST>_<DTR>_<4B>
        # values in <> are optional
        return self.name.split("_")[1:]

    @property
    def cmd_width(self):
        return self._get_configs[0]

    @property
    def addr_width(self):
        return self._get_configs[1]

    @property
    def data_width(self):
        return self._get_configs[2]

    def name_from_value(self, v):
        for (name, value) in self.__dict__.items():
            if value == v:
                return name



from litespi.spi_nor_flash_module import SpiNorFlashModule
from litespi.opcodes import SpiNorFlashOpCodes
from litespi.ids import SpiNorFlashManufacturerIDs

# Define non-generated SPI NOR chips here

class MX25L12833F(SpiNorFlashModule):
    """MX25L12833F

    Datasheet: https://www.macronix.com/Lists/Datasheet/Attachments/7447/MX25L12833F,%203V,%20128Mb,%20v1.0.pdf
    """

    manufacturer_id = SpiNorFlashManufacturerIDs.MACRONIX
    device_id = 0x2018
    name = "mx25l12833f"

    total_size  =   16777216   # bytes
    page_size   =        256   # bytes
    total_pages =      65536

    supported_opcodes = [
        SpiNorFlashOpCodes.READ_1_1_1,
        SpiNorFlashOpCodes.READ_1_1_1_FAST,
        SpiNorFlashOpCodes.READ_1_1_2,
        SpiNorFlashOpCodes.READ_1_2_2,
        SpiNorFlashOpCodes.READ_1_1_4,
        SpiNorFlashOpCodes.READ_1_4_4,
        SpiNorFlashOpCodes.READ_4_4_4,
        SpiNorFlashOpCodes.PP_1_1_1,
        SpiNorFlashOpCodes.PP_1_4_4,
    ]
    dummy_bits = 8


class MX25L6406E(SpiNorFlashModule):
    """MX25L6406E

    Datasheet: https://www.macronix.com/Lists/Datasheet/Attachments/7370/MX25L6406E,%203V,%2064Mb,%20v1.9.pdf
    """

    manufacturer_id = SpiNorFlashManufacturerIDs.MACRONIX
    device_id = 0x2017
    name = "mx25l6406e"

    total_size  =    8388608   # bytes
    page_size   =        256   # bytes
    total_pages =      32768

    supported_opcodes = [
        SpiNorFlashOpCodes.READ_1_1_1,
        SpiNorFlashOpCodes.READ_1_1_1_FAST,
        SpiNorFlashOpCodes.READ_1_1_2,
        SpiNorFlashOpCodes.PP_1_1_1,
    ]
    dummy_bits = 8


class W25Q512JV(SpiNorFlashModule):
    """W25Q512JV

    Datasheet: https://www.winbond.com/resource-files/W25Q512JV%20SPI%20RevB%2006252019%20KMS.pdf
    """

    manufacturer_id = SpiNorFlashManufacturerIDs.WINBOND
    device_id = 0x4021
    name = "w25q512jv"

    total_size  =   67108864   # bytes
    page_size   =        256   # bytes
    total_pages =     262144

    supported_opcodes = [
        SpiNorFlashOpCodes.READ_1_1_1,
        SpiNorFlashOpCodes.READ_1_1_1_4B,
        SpiNorFlashOpCodes.READ_1_1_1_FAST,
        SpiNorFlashOpCodes.READ_1_1_1_FAST_4B,
        SpiNorFlashOpCodes.READ_1_1_2,
        SpiNorFlashOpCodes.READ_1_1_2_4B,
        SpiNorFlashOpCodes.READ_1_2_2,
        SpiNorFlashOpCodes.READ_1_2_2_4B,
        SpiNorFlashOpCodes.READ_1_1_4,
        SpiNorFlashOpCodes.READ_1_1_4_4B,
        SpiNorFlashOpCodes.READ_1_4_4,
        SpiNorFlashOpCodes.READ_1_4_4_4B,
        SpiNorFlashOpCodes.PP_1_1_1,
        SpiNorFlashOpCodes.PP_1_1_1_4B,
        SpiNorFlashOpCodes.PP_1_1_4,
        SpiNorFlashOpCodes.PP_1_1_4_4B,
    ]
    dummy_bits = 8


class W74M64FV(SpiNorFlashModule):
    """W74M64FV

    Datasheet: https://www.winbond.com/resource-files/w74m64fv%20revb.pdf
    """

    manufacturer_id = SpiNorFlashManufacturerIDs.WINBOND
    device_id = 0x0000 # FIXME
    name = "w74m64fv"

    total_size  =    8388608   # bytes
    page_size   =        256   # bytes
    total_pages =      32768

    supported_opcodes = [
        SpiNorFlashOpCodes.READ_1_1_1,
        SpiNorFlashOpCodes.READ_1_1_4,
        SpiNorFlashOpCodes.PP_1_1_1,
    ]
    dummy_bits = 8



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

from litex.gen.genlib.misc import WaitTimer

from litespi.common import *

from litex.soc.interconnect.csr import *

from litex.soc.integration.doc import AutoDoc

from litespi.phy.generic_sdr import LiteSPISDRPHYCore
from litespi.phy.generic_ddr import LiteSPIDDRPHYCore

# LiteSPI PHY --------------------------------------------------------------------------------------

class LiteSPIPHY(Module, AutoCSR, AutoDoc):
    """LiteSPI PHY instantiator

    The ``LiteSPIPHY`` class instantiate generic PHY - ``LiteSPIPHYCore`` that can be connected to the ``LiteSPICore``,
    handles optional clock domain wrapping for whole PHY and interfaces streams and CS signal from PHY logic.

    Parameters
    ----------
    pads : Object
        SPI pads description.

    flash : SpiNorFlashModule
        SpiNorFlashModule configuration object.

    device : str
        Device type for use by the ``LiteSPIClkGen``.

    clock_domain : str
        Name of LiteSPI clock domain.

    default_divisor : int (1:1 rate)
        Default frequency divisor for clkgen.

    rate : str
        Rate: 1:1 SDR, 1:2 DDR.

    extra_latency : int (1:2 rate)
        Compensate for additional DDROutput/Input latency

    Attributes
    ----------
    source : Endpoint(spi_phy2core_layout), out
        Data stream from ``LiteSPIPHYCore``.

    sink : Endpoint(spi_core2phy_layout), in
        Control stream from ``LiteSPIPHYCore``.

    cs : Signal(), in
        Flash CS signal from ``LiteSPIPHYCore``.
    """

    def __init__(self, pads, flash, device="xc7", clock_domain="sys", default_divisor=9, cs_delay=10, rate="1:1", extra_latency=0):
        assert rate in ["1:1", "1:2"]
        if rate == "1:1":
            self.phy = LiteSPISDRPHYCore(pads, flash, device, clock_domain, default_divisor, cs_delay)
        if rate == "1:2":
            self.phy = LiteSPIDDRPHYCore(pads, flash, cs_delay, extra_latency)

        self.flash = flash

        self.source = self.phy.source
        self.sink   = self.phy.sink
        self.cs     = self.phy.cs

        # # #

        if clock_domain != "sys":
            self.phy = ClockDomainsRenamer(clock_domain)(self.phy)

        self.submodules.spiflash_phy = self.phy

    def get_csrs(self):
        return self.spiflash_phy.get_csrs()



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import enum


class CFIManufacturerIDs(enum.Enum):
    """Manufacturer IDs from the CFI standard.

    Common Flash Interface (CFI) is a standard introduced by the Joint Electron
    Device Engineering Council (JEDEC) to allow in-system or programmer reading
    of flash device characteristics, which is equivalent to having data sheet
    parameters located in the device.
    """
    AMD      = 0x0001
    AMIC     = 0x0037
    ATMEL    = 0x001F
    EON      = 0x001C
    ESMT     = 0x008C
    FUJITSU  = 0x0004
    HYUNDAI  = 0x00AD
    INTEL    = 0x0089
    ISSI     = 0x009D
    MACRONIX = 0x00C2
    NEC      = 0x0010
    PMC      = 0x009D
    SAMSUNG  = 0x00EC
    SANYO    = 0x0062
    SHARP    = 0x00B0
    SST      = 0x00BF
    ST       = 0x0020 # STMicroelectronics
    MICRON   = 0x002C
    THOMSON  = 0x00BA
    TOSHIBA  = 0x0098
    WINBOND  = 0x00DA


class SpiNorFlashManufacturerIDs(enum.Enum):
    """Manufacturer IDs for SPI NOR flash chips.

    The first byte returned from the flash after sending opcode SPINor_OP_RDID.
    Sometimes these are the same as CFI IDs, but sometimes they aren't.
    """
    AMIC       = CFIManufacturerIDs.AMIC.value
    ATMEL      = CFIManufacturerIDs.ATMEL.value
    EON        = CFIManufacturerIDs.EON.value
    ESMT       = CFIManufacturerIDs.ESMT.value
    FUJITSU    = CFIManufacturerIDs.FUJITSU.value
    GIGADEVICE = 0xc8
    INTEL      = CFIManufacturerIDs.INTEL.value
    ISSI       = CFIManufacturerIDs.ISSI.value
    ST         = CFIManufacturerIDs.ST.value
    MICRON     = CFIManufacturerIDs.MICRON.value
    MACRONIX   = CFIManufacturerIDs.MACRONIX.value
    SPANSION   = CFIManufacturerIDs.AMD.value
    SANYO      = CFIManufacturerIDs.SANYO.value
    SST        = CFIManufacturerIDs.SST.value
    THOMSON    = CFIManufacturerIDs.THOMSON.value
    WINBOND    = 0xef # Also used by some Spansion

    NONJEDEC   = 0x0



#!/usr/bin/env python3

from setuptools import setup
from setuptools import find_packages

with open("README.md", "r", encoding="utf-8") as fp:
    long_description = fp.read()

setup(
    name                 = "litespi",
    description          = "Small footprint and configurable SPI core",
    version                       = "2023.12",
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    author               = "LiteSPI Developers",
    url                  = "https://github.com/litex-hub",
    download_url         = "https://github.com/litex-hub/litespi",
    test_suite           = "test",
    license              = "BSD",
    python_requires      = "~=3.7",
    packages             = find_packages(exclude=("test*", "sim*", "doc*", "examples*")),
    include_package_data = True,
    entry_points         = {
        "console_scripts": [
            "litespi_gen=litespi.gen:main",
        ],
    },
)



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen.genlib.misc import WaitTimer

from litex.soc.interconnect import wishbone, stream
from litex.gen.common import reverse_bytes
from litex.soc.interconnect.csr import *

from litespi.common import *
from migen.genlib.cdc import MultiReg

cmd_oe_mask  = {
    1: 0b00000001,
    2: 0b00000011,
    4: 0b00001111,
    8: 0b11111111,
}
addr_oe_mask = {
    1: 0b00000001,
    2: 0b00000011,
    4: 0b00001111,
    8: 0b11111111,
}

class LiteSPIMMAP(Module, AutoCSR):
    """Memory-mapped SPI Flash controller.

    The ``LiteSPIMMAP`` class provides a Wishbone slave that must be connected to a LiteSPI PHY.

    It supports sequential accesses so that command and address is only sent when necessary.

    Parameters
    ----------
    endianness : string
        If endianness is set to ``little`` then byte order of each 32-bit word coming from flash will be reversed.

    Attributes
    ----------
    source : Endpoint(spi_core2phy_layout), out
        PHY control interface.

    sink : Endpoint(spi_phy2core_layout), in
        PHY data interface.

    bus : Interface(), out
        Wishbone interface for memory-mapped flash access.

    cs : Signal(), out
        CS signal for the flash chip, should be connected to cs signal of the PHY.

    dummy_bits : CSRStorage
        Register which hold a number of dummy bits to send during transmission.
    """
    def __init__(self, flash, clock_domain="sys", endianness="big", with_csr=True):
        self.source = source = stream.Endpoint(spi_core2phy_layout)
        self.sink   = sink   = stream.Endpoint(spi_phy2core_layout)
        self.bus    = bus    = wishbone.Interface()
        self.cs     = cs     = Signal()

        # Burst Control.
        burst_cs      = Signal()
        burst_adr     = Signal(len(bus.adr), reset_less=True)
        burst_timeout = WaitTimer(MMAP_DEFAULT_TIMEOUT)
        self.submodules += burst_timeout

        cmd_bits  = 8
        data_bits = 32

        if flash.cmd_width == 1:
            self._default_dummy_bits = flash.dummy_bits if flash.fast_mode else 0
        elif flash.cmd_width == 4:
            self._default_dummy_bits = flash.dummy_bits * 3 if flash.fast_mode else 0
        else:
            raise NotImplementedError(f'Command width of {flash.cmd_width} bits is currently not supported!')

        self._spi_dummy_bits = spi_dummy_bits = Signal(8)

        if with_csr:
            self.dummy_bits = dummy_bits = CSRStorage(8, reset=self._default_dummy_bits)
            if clock_domain != "sys":
                self.specials += MultiReg(dummy_bits.storage, spi_dummy_bits, clock_domain)
            else:
                self.comb += spi_dummy_bits.eq(dummy_bits.storage)
        else:
            self.comb += spi_dummy_bits.eq(self._default_dummy_bits)

        dummy = Signal(data_bits, reset=0xdead)

        # FSM.
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            # Keep CS active after Burst for Timeout.
            burst_timeout.wait.eq(1),
            NextValue(burst_cs, burst_cs & ~burst_timeout.done),
            cs.eq(burst_cs),
            # On Bus Read access...
            If(bus.cyc & bus.stb & ~bus.we,
                # If CS is still active and Bus address matches previous Burst address:
                # Just continue the current Burst.
                If(burst_cs & (bus.adr == burst_adr),
                    NextState("BURST-REQ")
                # Otherwise initialize a new Burst.
                ).Else(
                    cs.eq(0),
                    NextState("BURST-CMD")
                )
            )
        )

        fsm.act("BURST-CMD",
            cs.eq(1),
            source.valid.eq(1),
            source.data.eq(flash.read_opcode.code), # send command.
            source.len.eq(cmd_bits),
            source.width.eq(flash.cmd_width),
            source.mask.eq(cmd_oe_mask[flash.cmd_width]),
            NextValue(burst_adr, bus.adr),
            If(source.ready,
                NextState("CMD-RET"),
            )
        )

        fsm.act("CMD-RET",
            cs.eq(1),
            sink.ready.eq(1),
            If(sink.valid,
                NextState("BURST-ADDR"),
            )
        )

        fsm.act("BURST-ADDR",
            cs.eq(1),
            source.valid.eq(1),
            source.width.eq(flash.addr_width),
            source.mask.eq(addr_oe_mask[flash.addr_width]),
            source.data.eq(Cat(Signal(2), bus.adr)), # send address.
            source.len.eq(flash.addr_bits),
            NextValue(burst_cs, 1),
            NextValue(burst_adr, bus.adr),
            If(source.ready,
                NextState("ADDR-RET"),
            )
        )

        fsm.act("ADDR-RET",
            cs.eq(1),
            sink.ready.eq(1),
            If(sink.valid,
                If(spi_dummy_bits == 0,
                    NextState("BURST-REQ"),
                ).Else(
                    NextState("DUMMY"),
                )
            )
        )

        fsm.act("DUMMY",
            cs.eq(1),
            source.valid.eq(1),
            source.width.eq(flash.addr_width),
            source.mask.eq(addr_oe_mask[flash.addr_width]),
            source.data.eq(dummy),
            source.len.eq(spi_dummy_bits),
            If(source.ready,
                NextState("DUMMY-RET"),
            )
        )

        fsm.act("DUMMY-RET",
            cs.eq(1),
            sink.ready.eq(1),
            If(sink.valid,
                NextState("BURST-REQ"),
            )
        )

        fsm.act("BURST-REQ",
            cs.eq(1),
            source.valid.eq(1),
            source.last.eq(1),
            source.width.eq(flash.bus_width),
            source.len.eq(data_bits),
            source.mask.eq(0),
            If(source.ready,
                NextState("BURST-DAT"),
            )
        )

        fsm.act("BURST-DAT",
            cs.eq(1),
            sink.ready.eq(1),
            bus.dat_r.eq({"big": sink.data, "little": reverse_bytes(sink.data)}[endianness]),
            If(sink.valid,
                bus.ack.eq(1),
                NextValue(burst_adr, burst_adr + 1),
                NextState("IDLE"),
            )
        )



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause


from litespi.spi_nor_features import SpiNorFeatures
from litespi.opcodes import SpiNorFlashOpCodes
from litespi.spi_nor_flash_metasizes import MetaSizes


class SpiNorFlashModule(metaclass=MetaSizes):
    """SPI NOR flash module

    The ``SpiNorFlashModule`` class provides an essential information about chip capabilities
    along with supported commands.

    Parameters
    ----------
    default_read_cmd : SpiNorFlashOpCode
        Default user defined read command also used to
        configure a chip into initial desired mode.

    read_cmds : list of SpiNorFlashOpCode
        List of read commands that can be used by SPI NOR controller
        default: default_read_cmd

    program_cmd : SpiNorFlashOpCode
        User defined program command.
        default: PP_1_1_1 (single mode page program)

    erase_cmd : SpiNorFlashOpCode
        User defined erase command. Default value is Sector Erase.
        default: SE (sector erase)

    Attributes
    ----------
    manufacturer_id : SpiNorFlashManufacturerIDs
        ID of a vendor

    device_id : int
        ID of a chip

    name : str
        name of a chip

    total_size : int
        total size of a chip in bytes

    page_size : int
        page size of a chip in bytes

    total_pages : int
        chip's total number of pages

    addr_width : int
        number of lines used to send address

    addr_bits : int
        number of address bits (24 or 32)

    supported_opcodes : list of SpiNorFlashOpCode
        list of supported opcodes (READ/PP)

    dummy_bits : int
        number of dummy bits to send needed for a correct transfer

    fast_mode : bool
        chip configured in fast mode
        True -> dummy_bits used
        False -> no dummy_bits

    cmd_width : int
        number of lines used to send command

    read_cmds : list of SpiNorFlashOpCode
        list of available read commands

    program_opcode : SpiNorFlashOpCode
        command used to program the memory

    erase_opcode : SpiNorFlashOpCode
        command used to erase the memory

    bus_width : int
        expected number of data lines

    ddr : bool
        setup module to run in DDR mode

    Properties
    ----------
    read_opcode : SpiNorFlashOpCode
        command used to read the memory
        as well as configures chip into desired mode
    """

    @property
    def read_opcode(self):
        return self._read_opcode

    @read_opcode.setter
    def read_opcode(self, read_cmd):
        """read opcode

        Property ``read_opcode`` defines chip mode,
        hence follwing attributes are set:

            * cmd_width
            * addr_width
            * bus_width
            * addr_bits
            * ddr
        """

        cfg = SpiNorFlashOpCodes()
        cfg.name = SpiNorFlashOpCodes.name_from_value(SpiNorFlashOpCodes, read_cmd)

        self.ddr = "DTR" in cfg.name
        if "4B" in cfg.name:
            self.addr_bits = 32
        else:
            self.addr_bits = 24


        self.cmd_width  = int(cfg.cmd_width)
        self.addr_width = int(cfg.addr_width)
        self.bus_width  = int(cfg.data_width)

        self.fast_mode = "FAST" in cfg.name or self.bus_width > 1

        self._read_opcode = read_cmd

    def check_bus_width(self, width):
        return width >= self.bus_width

    def _configure_chip(self, default_read_cmd,
                        erase_cmd,
                        program_cmd):
        # Check if commands on the list are supported
        for cmd in self.read_cmds:
            if cmd not in self.supported_opcodes:
                raise ValueError("""\
Read command (%s) not supported in chip %s!""" % (str(cmd), self.name))

        # Check if program command is supported
        if program_cmd not in self.supported_opcodes:
            raise ValueError("Read command (%s) not supported in chip %s!" % (str(read_cmd), self.name))

        # Set commands
        self.read_opcode = default_read_cmd
        self.program_opcode = program_cmd
        self.erase_opcode = erase_cmd

    def __init__(self, default_read_cmd,
                 read_cmds=[],
                 program_cmd=SpiNorFlashOpCodes.PP_1_1_1,
                 erase_cmd=SpiNorFlashOpCodes.SE):
        # Check if mandatory attributes are set by an inheritor
        assert hasattr(self, 'manufacturer_id')
        assert hasattr(self, 'device_id')
        assert hasattr(self, 'name')
        assert hasattr(self, 'total_size')
        assert hasattr(self, 'page_size')
        assert hasattr(self, 'total_pages')
        assert hasattr(self, 'supported_opcodes')
        assert hasattr(self, 'dummy_bits')

        # Make sure default read command is on the list
        if default_read_cmd not in read_cmds:
            read_cmds.append(default_read_cmd)

        self.read_cmds = read_cmds

        # Configure a chip using provided default_read_cmd
        self._configure_chip(default_read_cmd,
                             erase_cmd,
                             program_cmd)



#
# This file is part of LiteSPI
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# Copyright (c) 2021 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen.genlib.misc import WaitTimer

from litespi.common import *
from litespi.clkgen import LiteSPIClkGen

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import *

from litex.build.io import SDROutput, SDRInput, SDRTristate

from litex.soc.integration.doc import AutoDoc

# LiteSPI PHY Core ---------------------------------------------------------------------------------

class LiteSPISDRPHYCore(Module, AutoCSR, AutoDoc):
    """LiteSPI PHY instantiator

    The ``LiteSPIPHYCore`` class provides a generic PHY that can be connected to the ``LiteSPICore``.

    It supports single/dual/quad/octal output reads from the flash chips.

    The following diagram shows how each clock configuration option relates to outputs and input sampling in DDR mode:

    .. wavedrom:: ../../doc/ddr-timing-diagram.json

    Parameters
    ----------
    pads : Object
        SPI pads description.

    flash : SpiNorFlashModule
        SpiNorFlashModule configuration object.

    device : str
        Device type for use by the ``LiteSPIClkGen``.

    default_divisor : int
        Default frequency divisor for clkgen.

    Attributes
    ----------
    source : Endpoint(spi_phy2core_layout), out
        Data stream.

    sink : Endpoint(spi_core2phy_layout), in
        Control stream.

    cs : Signal(), in
        Flash CS signal.

    clk_divisor : CSRStorage
        Register which holds a clock divisor value applied to clkgen.
    """
    def __init__(self, pads, flash, device, clock_domain, default_divisor, cs_delay):
        self.source           = source = stream.Endpoint(spi_phy2core_layout)
        self.sink             = sink   = stream.Endpoint(spi_core2phy_layout)
        self.cs               = Signal()
        self._spi_clk_divisor = spi_clk_divisor = Signal(8)

        self._default_divisor = default_divisor

        self.clk_divisor      = clk_divisor = CSRStorage(8, reset=self._default_divisor)

        # # #

        # Resynchronize CSR Clk Divisor to LiteSPI Clk Domain.
        self.submodules += ResyncReg(clk_divisor.storage, spi_clk_divisor, clock_domain)

        # Determine SPI Bus width and DQs.
        if hasattr(pads, "mosi"):
            bus_width = 1
        else:
            bus_width = len(pads.dq)
        assert bus_width in [1, 2, 4, 8]

        # Check if number of pads matches configured mode.
        assert flash.check_bus_width(bus_width)

        self.addr_bits  = addr_bits  = flash.addr_bits
        self.cmd_width  = cmd_width  = flash.cmd_width
        self.addr_width = addr_width = flash.addr_width
        self.data_width = data_width = flash.bus_width
        self.ddr        = ddr        = flash.ddr

        self.command = command = flash.read_opcode.code

        # Clock Generator.
        self.submodules.clkgen = clkgen = LiteSPIClkGen(pads, device, with_ddr=ddr)
        self.comb += [
            clkgen.div.eq(spi_clk_divisor),
            clkgen.sample_cnt.eq(1),
            clkgen.update_cnt.eq(1),
        ]

        # CS control.
        cs_timer  = WaitTimer(cs_delay + 1) # Ensure cs_delay cycles between XFers.
        cs_enable = Signal()
        self.submodules += cs_timer
        self.comb += cs_timer.wait.eq(self.cs)
        self.comb += cs_enable.eq(cs_timer.done)
        self.comb += pads.cs_n.eq(~cs_enable)

        # I/Os.
        data_bits = 32
        cmd_bits  = 8

        if hasattr(pads, "mosi"):
            dq_o  = Signal()
            dq_i  = Signal(2)
            dq_oe = Signal() # Unused.
            self.specials += SDROutput(
                i = dq_o,
                o = pads.mosi
            )
            self.specials += SDRInput(
                i = pads.miso,
                o = dq_i[1]
            )
        else:
            dq_o  = Signal(len(pads.dq))
            dq_i  = Signal(len(pads.dq))
            dq_oe = Signal(len(pads.dq))
            for i in range(len(pads.dq)):
                self.specials += SDRTristate(
                    io = pads.dq[i],
                    o  = dq_o[i],
                    oe = dq_oe[i],
                    i  = dq_i[i],
                )

        # Data Shift Registers.
        sr_cnt       = Signal(8, reset_less=True)
        sr_out_load  = Signal()
        sr_out_shift = Signal()
        sr_out       = Signal(len(sink.data), reset_less=True)
        sr_in_shift  = Signal()
        sr_in        = Signal(len(sink.data), reset_less=True)

        # Data Out Generation/Load/Shift.
        self.comb += [
            dq_oe.eq(sink.mask),
            Case(sink.width, {
                1 : dq_o.eq(sr_out[-1:]),
                2 : dq_o.eq(sr_out[-2:]),
                4 : dq_o.eq(sr_out[-4:]),
                8 : dq_o.eq(sr_out[-8:]),
            })
        ]
        self.sync += If(sr_out_load,
            sr_out.eq(sink.data << (len(sink.data) - sink.len))
        )
        self.sync += If(sr_out_shift,
            Case(sink.width, {
                1 : sr_out.eq(Cat(Signal(1), sr_out)),
                2 : sr_out.eq(Cat(Signal(2), sr_out)),
                4 : sr_out.eq(Cat(Signal(4), sr_out)),
                8 : sr_out.eq(Cat(Signal(8), sr_out)),
            })
        )

        # Data In Shift.
        self.sync += If(sr_in_shift,
            Case(sink.width, {
                1 : sr_in.eq(Cat(dq_i[1], sr_in)),
                2 : sr_in.eq(Cat(dq_i[:2], sr_in)),
                4 : sr_in.eq(Cat(dq_i[:4], sr_in)),
                8 : sr_in.eq(Cat(dq_i[:8], sr_in)),
            })
        )

        # FSM.
        self.submodules.fsm = fsm = FSM(reset_state="WAIT-CMD-DATA")
        fsm.act("WAIT-CMD-DATA",
            # Wait for CS and a CMD from the Core.
            If(cs_enable & sink.valid,
                # Load Shift Register Count/Data Out.
                NextValue(sr_cnt, sink.len - sink.width),
                sr_out_load.eq(1),
                # Start XFER.
                NextState("XFER"),
            ),
        )
        fsm.act("XFER",
            # Generate Clk.
            self.clkgen.en.eq(1),

            # Data In Shift.
            If(clkgen.posedge_reg2, sr_in_shift.eq(1)),

            # Data Out Shift.
            If(clkgen.negedge, sr_out_shift.eq(1)),

            # Shift Register Count Update/Check.
            If(self.clkgen.negedge,
                NextValue(sr_cnt, sr_cnt - sink.width),
                # End XFer.
                If(sr_cnt == 0,
                    NextState("XFER-END"),
                ),
            ),

        )
        fsm.act("XFER-END",
            # Last data already captured in XFER when divisor > 0 so only capture for divisor == 0.
            If((spi_clk_divisor > 0) | clkgen.posedge_reg2,
                # Accept CMD.
                sink.ready.eq(1),
                # Capture last data (only for spi_clk_divisor == 0).
                sr_in_shift.eq(spi_clk_divisor == 0),
                # Send Status/Data to Core.
                NextState("SEND-STATUS-DATA"),
            )
        )
        self.comb += source.data.eq(sr_in)
        fsm.act("SEND-STATUS-DATA",
            # Send Data In to Core and return to WAIT when accepted.
            source.valid.eq(1),
            source.last.eq(1),
            If(source.ready,
                NextState("WAIT-CMD-DATA"),
            )
        )



#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause


import enum

class SpiNorFeatures(enum.Flag):
    FEATURE_ERASED_ZERO     = enum.auto()
    """Erasing flash ends up with 0x00 rather than 0xff."""

    FEATURE_NO_ERASE        = enum.auto()
    """Chip does not support erase commands."""

    FEATURE_OTP             = enum.auto()
    """Has a one time programming area."""

    FEATURE_WRSR_EWSR       = enum.auto()
    """Before sending the WRSR instruction the Enable-Write-Status-Register must be written too."""

    FEATURE_WRSR_WREN       = enum.auto()
    """Before sending WRSR instruction, the Write Enable (WREN) instruction must be decoded and executed to set the Write Enable Latch (WEL) bit in advance."""

    FEATURE_WRSR_EITHER     = FEATURE_WRSR_EWSR | FEATURE_WRSR_WREN

    FEATURE_QPI             = enum.auto()
    """Supports Quad-SPI-mode - ???"""

    FEATURE_4BA_ENTER       = enum.auto()
    """Can enter/exit 4BA mode with instructions 0xb7/0xe9 w/o WREN."""

    FEATURE_4BA_ENTER_WREN  = enum.auto()
    """Can enter/exit 4BA mode with instructions 0xb7/0xe9 after WREN."""

    FEATURE_4BA_ENTER_EAR7  = enum.auto()
    """Can enter/exit 4BA mode by setting bit7 of the ext addr reg."""

    FEATURE_4BA_EXT_ADDR    = enum.auto()
    """Regular 3-byte operations can be used by writing the most significant address byte into an extended address register."""

    FEATURE_4BA_READ        = enum.auto()
    """Native 4BA read instruction (0x13) is supported."""

    FEATURE_4BA_FAST_READ   = enum.auto()
    """Native 4BA fast read instruction (0x0c) is supported."""

    FEATURE_4BA_WRITE       = enum.auto()
    """Native 4BA byte program (0x12) is supported."""

    # 4BA Shorthands
    FEATURE_4BA_NATIVE = (FEATURE_4BA_READ       | FEATURE_4BA_FAST_READ | FEATURE_4BA_WRITE)
    FEATURE_4BA        = (FEATURE_4BA_ENTER      | FEATURE_4BA_EXT_ADDR  | FEATURE_4BA_NATIVE)
    FEATURE_4BA_WREN   = (FEATURE_4BA_ENTER_WREN | FEATURE_4BA_EXT_ADDR  | FEATURE_4BA_NATIVE)
    FEATURE_4BA_EAR7   = (FEATURE_4BA_ENTER_EAR7 | FEATURE_4BA_EXT_ADDR  | FEATURE_4BA_NATIVE)




#
# This file is part of LiteSPI
#
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.integration.doc import AutoDoc, ModuleDoc

from litex.build.io import SDROutput, DDROutput

class DDRLiteSPIClkGen(Module, AutoDoc):
    """SPI Clock generator

    The ``DDRLiteSPIClkGen`` class provides a generic SPI clock generator.

    The class can be combined with DDR PHY Core.

    Parameters
    ----------
    pads : Object
        SPI pads description.

    Attributes
    ----------
    en : Signal(), in
        Clock enable input, output clock will be generated if set to 1, 0 resets the core.
    """
    def __init__(self, pads):
        self.en         = en         = Signal()

        self.specials += DDROutput(i1=en, i2=0, o=pads.clk)


class LiteSPIClkGen(Module, AutoDoc):
    """SPI Clock generator

    The ``LiteSPIClkGen`` class provides a generic SPI clock generator.

    It supports accessing CLK pin on a reserved pin by instantiating device specific modules (currently 7 Series only).

    Parameters
    ----------
    pads : Object
        SPI pads description.

    device : str
        Device type for determining how to get output pin if it was not provided in pads.

    cnt_width : int
        Width of the internal counter ``cnt`` used for dividing the clock.

    with_ddr : bool
        Generate additional ``sample`` and ``update`` signals.

    Attributes
    ----------
    div : Signal(8), in
        Clock divisor, output clock frequency will be equal to ``sys_clk_freq/(2*(1+div))``.

    posedge : Signal(), out
        Outputs 1 when there is a rising edge on the generated clock, 0 otherwise.

    negedge : Signal(), out
        Outputs 1 when there is a falling edge on the generated clock, 0 otherwise.

    en : Signal(), in
        Clock enable input, output clock will be generated if set to 1, 0 resets the core.

    sample : Signal(), out
        Outputs 1 when ``sample_cnt==cnt``, can be used to sample incoming DDR data.

    sample_cnt : Signal(8), in
        Controls generation of the ``sample`` signal.

    update : Signal(), out
        Outputs 1 when ``update_cnt==cnt``, can be used to update outgoing DDR data.

    update_cnt : Signal(8), in
        Controls generation of the ``update`` signal.
    """
    def __init__(self, pads, device, cnt_width=8, with_ddr=False):
        self.div        = div        = Signal(cnt_width)
        self.sample_cnt = sample_cnt = Signal(cnt_width)
        self.update_cnt = update_cnt = Signal(cnt_width)
        self.posedge    = posedge    = Signal()
        self.negedge    = negedge    = Signal()
        self.sample     = sample     = Signal()
        self.update     = update     = Signal()
        self.en         = en         = Signal()
        cnt             = Signal(cnt_width)
        en_int          = Signal()
        clk             = Signal()

        self.comb += [
            posedge.eq(en & ~clk & (cnt == div)),
            negedge.eq(en & clk & (cnt == div)),
            sample.eq(cnt == sample_cnt),
            update.eq(cnt == update_cnt),
        ]

        # Delayed edge to account for IO register delays.
        self.posedge_reg  = posedge_reg  = Signal()
        self.posedge_reg2 = posedge_reg2 = Signal()

        self.sync += [
            posedge_reg.eq(posedge),
            posedge_reg2.eq(posedge_reg),
        ]

        self.sync += [
            If(en | en_int,
                If(cnt < div,
                    cnt.eq(cnt+1),
                ).Else(
                    cnt.eq(0),
                    clk.eq(~clk),
                )
            ).Else(
                clk.eq(0),
                cnt.eq(0),
            )
        ]

        if not hasattr(pads, "clk"):
            # Clock output needs to be registered like an SDROutput.
            clk_reg = Signal()
            self.sync += clk_reg.eq(clk)

            if device.startswith("xc7"):
                cycles = Signal(4)
                self.specials += Instance("STARTUPE2",
                    i_CLK       = 0,
                    i_GSR       = 0,
                    i_GTS       = 0,
                    i_KEYCLEARB = 0,
                    i_PACK      = 0,
                    i_USRCCLKO  = clk_reg,
                    i_USRCCLKTS = 0,
                    i_USRDONEO  = 1,
                    i_USRDONETS = 1,
                )
                # startupe2 needs 3 usrcclko cycles to switch over to user clock
                self.comb += en_int.eq(cycles < 3)
                self.sync += If(en_int & posedge, cycles.eq(cycles+1))
            elif device.startswith("LFE5U"):
                self.specials += Instance("USRMCLK",
                    i_USRMCLKI  = clk_reg,
                    i_USRMCLKTS = 0
                )
            else:
                raise NotImplementedError
        else:
            self.specials += SDROutput(i=clk, o=pads.clk)
